/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "s_problem.hpp"

//**************************************************************************

bool SomaProb::check_solution(const unsigned int num_pieces,
                              const std::vector<SomaPiece> *const pieces,
                              const unsigned int num_blocks,
                              const std::list<SomaVector<int> > *const blocks,
                              const SomaSolution *solution){
  
  SomaVector<int> pos,abc,def;
  unsigned int ori,found;
  SomaPiece pc(1);
  std::list<SomaVector<int> >::const_iterator it;
  
  // compare number of blocks
  unsigned int sum = 0;
  for(unsigned int i=0;i<num_pieces;i++){
    if(solution->get(i,&pos,&ori)){
      sum += (*pieces)[i].get_num_blocks();
    }
  }
  if(sum != num_blocks) return false;
    
  // trying to find all blocks in solution
  it = blocks->begin();
  while(it!= blocks->end()){
    
    // look to which piece this block belongs
    found = 0;
    for(unsigned int i=0;i<num_pieces;i++){
      if(solution->get(i,&pos,&ori)){
        pc.set((*pieces)[i]);
        pc.set_orientation(ori);
        abc = *it - pos;
        
        for(unsigned int j=0;j<pc.get_num_blocks();j++){
          pc.get(j,&def);
          
          if(def == abc) found++;
        }
        
      }
    }
    
    //whoops .. not found exactly once
    if (found !=1 ) return false;
    
    it++;
  }
  
  //try to find all blocks from solution in block list
  for(unsigned int i=0;i<num_pieces;i++){
    if(solution->get(i,&pos,&ori)){
      pc.set((*pieces)[i]);
      pc.set_orientation(ori);
      
      for(unsigned int j=0;j<pc.get_num_blocks();j++){
        pc.get(j,&def);
        def += pos;
        
        found = 0;
        it = blocks->begin();
        while(it != blocks->end()){
          if(*it == def) found++;
          it++;
        }
        
        //not found exactly once
        if(found != 1) return false;
        
      }
      
    }
  }
  
  
  return true;
}

//**************************************************************************

SomaProb::SomaProb(){
  pieces = 0;
  num_pieces = 0;
  solution = 0;
  blocks = 0;
  num_blocks = 0;
  max_blocks = 0;
  iter = 0;
}

//**************************************************************************

SomaProb* SomaProb::create_problem(std::istream& strm){
  SomaProb* prb = new SomaProb();
  
  if(prb->read(strm)) return prb;
  else{
    delete prb;
    return 0;
  }
  
}

//**************************************************************************

SomaProb::~SomaProb(){
  if(pieces) delete pieces;
  if(blocks) delete blocks;
  if(solution) delete solution;
}

//**************************************************************************

const SomaPiece* SomaProb::get_piece(const unsigned int piece_num)const{
#ifdef SOMA_DEBUG
  if(piece_num >= num_pieces){
    std::cerr << "Wrong piece number in SomaProb::get_piece()." << std::endl;
    return &(*pieces)[0];
  }
#endif

  return &(*pieces)[piece_num];  
}
  
//**************************************************************************

void SomaProb::add_solution(SomaSolution& solution_in){
#ifdef SOMA_DEBUG
  if(solution_in.get_num_pieces() != num_pieces){
    std::cerr << "Number of pieces does not match in SomaProb::set_solution()."
              << std::endl;
    return;
  }
#endif
  if(!solution) solution = new SomaSolutions();
  solution->addSolution(solution_in);
}
  
//**************************************************************************

const SomaSolutions* SomaProb::get_solution(){
  return solution;
}
  
//**************************************************************************

bool SomaProb::reset_blocks()const{
  iter = blocks->begin();
  return !(iter == blocks->end());
}

//**************************************************************************

bool SomaProb::get_block(SomaVector<int> *const xyz)const{
  if(iter != blocks->end()){
    *xyz = *iter;
    iter++;
    return !(iter == blocks->end());
  }
  return false;
}

//**************************************************************************

bool SomaProb::more_blocks()const{
  return !(iter == blocks->end());
}

//**************************************************************************

bool SomaProb::insert_block(const SomaVector<int>& xyz){
  if (solution) return false;
  
  if((num_blocks < max_blocks)&&(!is_block(xyz))){
    blocks->push_back(xyz);
    num_blocks++;
    return true;
  }else{
    return false;
  }
}

//**************************************************************************

bool SomaProb::remove_block(const SomaVector<int>& xyz){
  
  if (solution) return false;
  
  std::list<SomaVector<int> >::iterator it = blocks->begin();
  while(it != blocks->end()){
    if(*it == xyz){
      blocks->erase(it);
      num_blocks--;
      iter = blocks->begin();
      return true;
    }
    it++;
  }
  
  return false;
}

//**************************************************************************

void SomaProb::clear_blocks(){
  if (solution) return;
  
  std::list<SomaVector<int> >::iterator it = blocks->begin();
  while(it != blocks->end()){
    it = blocks->erase(it);
  }
  iter = blocks->begin();
  num_blocks = 0;
}

//**************************************************************************

bool SomaProb::is_block(const SomaVector<int>& xyz)const{
  
  std::list<SomaVector<int> >::const_iterator it = blocks->begin();
  while(it != blocks->end()){
    if(*it == xyz) return true;
    it++;
  }
  
  return false;
}

//**************************************************************************

SomaVector<double> SomaProb::center(){
  std::list<SomaVector<int> >::iterator it = blocks->begin();
  SomaVector<int> max,min;
  if(it != blocks->end()){
    min = max = *it;
    it++;
  }
  while(it != blocks->end()){
    if(min.x > (*it).x) min.x = (*it).x;
    if(min.y > (*it).y) min.y = (*it).y;
    if(min.z > (*it).z) min.z = (*it).z;

    if(max.x < (*it).x) max.x = (*it).x;
    if(max.y < (*it).y) max.y = (*it).y;
    if(max.z < (*it).z) max.z = (*it).z;
    
    it++;
  }
  
  min += max;
  SomaVector<double> retval(min);
  min /= 2;
  retval /= 2;
  retval -= SomaVector<double>(min);
  
  min.negate();
  retval.negate();
  
  it = blocks->begin();
  while(it != blocks->end()){
    *it += min;
    it++;
  }
  
  if(solution){
    for(unsigned int i=0;i<solution->size();i++){
      solution->get(i)->shift(min);
    }
  }
  
  return retval;
}

//**************************************************************************

void SomaProb::unsolve(){
  if(solution) delete solution;
  solution = 0;
}

//**************************************************************************

void SomaProb::write(std::ostream& strm)const{
  strm << " { (";
  strm.width(3);
  strm << num_pieces << ") " << std::endl;
  for(unsigned int i=0;i<num_pieces;i++){
    strm << (*pieces)[i];
  }
  
  if(num_blocks > 0){
    strm << " [ (";
    strm.width(3);
    strm << num_blocks << ") ";
    std::list<SomaVector<int> >::const_iterator it = blocks->begin();
    unsigned int count = 0;
    while(it!=blocks->end()){
      if(count%4 == 0) strm << std::endl;
      strm << *it;
      count++;
      it++;
    }
    strm << " ] "; 
  }
  
  if(solution){
    strm << std::endl << " <";
    strm << *solution;
    strm << " > " << std::endl;  
  }
  
  strm << " } " << std::endl;
}

//**************************************************************************

bool SomaProb::read(std::istream& strm){
  char S;
  bool error = false;
  
  unsigned int nw_num_pieces=0;
  std::vector<SomaPiece>* newpieces=0;
  
  SomaSolutions *newsolution=0;

  unsigned int nw_max_blocks=0;
  unsigned int nw_num_blocks=0;
  std::list<SomaVector<int> >* newblocks = new std::list<SomaVector<int> >;
  std::list<SomaVector<int> >::const_iterator it,jt;
  
  // read pieces 
  S=' '; 
  strm >> S;
  
  if(S!='{'){ 
    std::cerr << "Error in SomaProb::read(). '{' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S=='('){
    strm >> nw_num_pieces;
    
    if(nw_num_pieces > 0){
      newpieces = new std::vector<SomaPiece>(nw_num_pieces);
    }else{
      std::cerr << "Number of pieces 0 in SomaProb::read()." << std::endl;
      error = true;
      goto error_stop;
    }
  }else{
    std::cerr << "Error in SomaProb::read(). '(' expected" << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S!=')'){
    std::cerr << "Error in SomaProb::read(). ')' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  for(unsigned int i=0;i<nw_num_pieces;i++){
    if(!(*newpieces)[i].read(strm)){
      error = 1;
      goto error_stop;
    }
  }
  
  for(unsigned int i=0;i<nw_num_pieces;i++) 
    nw_max_blocks += (*newpieces)[i].get_num_blocks();
    
    S=' '; 
    strm >> S;
    
      if(S=='}'){
        goto error_stop;
      }else if(S!='['){
        std::cerr << "Error in SomaProb::read(). '}' or '[' expected." 
                  << std::endl;
        error = true;
        goto error_stop;
      }
      
      // read problem 
      S=' '; 
      strm >> S;
      
      if(S=='('){
        strm >> nw_num_blocks;
        
        if(nw_num_blocks == 0 || nw_num_blocks > nw_max_blocks){
          std::cerr << "Incorrect number of blocks in SomaProb::read()."
                    << std::endl;
          error = true; 
          goto error_stop;
        }
      }else{
        std::cerr << "Error in SomaProb::read(). '(' expected." 
                  << std::endl;
        error = true;
        goto error_stop;
      }
      
      S=' '; 
      strm >> S;
      
      if(S!=')'){
        std::cerr << "Error in SomaProb::read(). ')' expected." << std::endl;
        error = true;
        goto error_stop;
      }
      
      for(unsigned int i=0;i<nw_num_blocks;i++){
        SomaVector<int> crd;
        if(crd.read(strm)){
          newblocks->push_back(crd);
        }else{
          error = true;
          goto error_stop;
        }
      }
      
      S=' '; 
      strm >> S;
      
      if(S!=']'){
        std::cerr << "Error in SomaProb::read(). ']' expected." << std::endl;
        error = true;
        goto error_stop;
      }
      
      it = newblocks->begin();
      
      while(it != newblocks->end()){
        jt = it; jt++;
        
        while(jt != newblocks->end()){
          
          if(*jt == *it){
            std::cerr << "Input sematically incorrect in SomaProb::read()." 
                      << std::endl;
            std::cerr << "No double blocks allowed." << std::endl;
            error = true; 
            goto error_stop;
          }
          
          jt++;
        }
        
        it++;
      }
      
      S=' '; 
      strm >> S;
      
  if(S=='}'){
    goto error_stop;
  }else if(S!='<'){
      std::cerr << "Error in SomaProb::read(). '}' or '<' expected." 
                << std::endl;
      error = true;
      goto error_stop;
  }
  
  newsolution = new SomaSolutions();
  if( !newsolution->read(strm) ){
    std::cerr << "Could not read solution in SomaProb::read()." << std::endl;
    error = true;
    goto error_stop;
  }
  
  // check that read solution is indeed a solution
  for(unsigned int i=0;i<newsolution->size();i++){
    if(!check_solution(nw_num_pieces,newpieces,nw_num_blocks,newblocks,newsolution->get(i))){
      std::cerr << "Error: given solution " << i+1 << "does not " 
                << "solve the problem in SomaProb::read()." << std::endl;
      error = true;
      goto error_stop; 
    }
  }
  
  
  S=' '; 
  strm >> S;
  
  if(S!='>'){
    std::cerr << "Error in SomaProb::read(). '>' expected." << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  S=' '; 
  strm >> S;
  
  if(S!='}'){
      std::cerr << "Error in SomaProb::read(). '}' expected." << std::endl;
      error = true;
      goto error_stop;
  }
  
  error_stop:
  
  if(error){
    
    if(newpieces) delete newpieces;
    if(newblocks) delete newblocks;
    if(newsolution) delete newsolution;
    
    return false;
    
  }else{
    
    if(pieces) delete pieces;
    if(blocks) delete blocks;
    if(solution) delete solution;
    
    solution = newsolution;
    
    pieces = newpieces;
    num_pieces = nw_num_pieces;

    blocks = newblocks;
    max_blocks = nw_max_blocks;
    num_blocks = nw_num_blocks;
    
    iter = blocks->begin();
    
    return true;
  }
}

//**************************************************************************
