/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "s_solution.hpp"
#include "s_piece.hpp"

//**************************************************************************

SomaSolution::SomaSolution(const unsigned int num_pieces_in):
                           num_pieces(num_pieces_in),
                           position(num_pieces),
                           orientation(num_pieces,SomaPiece::ORIENTATIONS){
  
}

//**************************************************************************

SomaSolution::SomaSolution(const SomaSolution& solution_in):
                           num_pieces(solution_in.num_pieces),
                           position(solution_in.position),
                           orientation(solution_in.orientation){

}

//**************************************************************************
  
SomaSolution::~SomaSolution(){

}

//**************************************************************************  

void SomaSolution::insert(const unsigned int piece,const SomaVector<int>& pos,
                          const unsigned int ori){
#ifdef SOMA_DEBUG
  if(piece >= num_pieces){
    std::cerr << "Piece out of range in SomaSolution::insert()" << std::endl;
    return;
  }
  if(ori >= SomaPiece::ORIENTATIONS){
    std::cerr << "Orientation out of range in SomaSolution::insert()"
              << std::endl;
    return;
  }
#endif

  position[piece] = pos;
  orientation[piece] = ori;
}

//**************************************************************************  

void SomaSolution::remove(const unsigned int piece){
#ifdef SOMA_DEBUG
  if(piece >= num_pieces){
    std::cerr << "Piece out of range in SomaSolution::remove()" << std::endl;
    return;
  }
#endif

  orientation[piece] = SomaPiece::ORIENTATIONS;
}

//**************************************************************************

void SomaSolution::clear(){
  for(unsigned int i=0;i<num_pieces;i++){
    orientation[i] = SomaPiece::ORIENTATIONS;
  }
}

//**************************************************************************  

void SomaSolution::shift(const SomaVector<int>& offset){
  for(unsigned int i=0;i<num_pieces;i++){
    if(orientation[i] != SomaPiece::ORIENTATIONS){
      position[i] += offset;
    }
  }
}

//**************************************************************************  

bool SomaSolution::get(const unsigned int piece,SomaVector<int> *const pos,
                       unsigned int *const ori)const{
#ifdef SOMA_DEBUG
  if(piece >= num_pieces){
    std::cerr << "Piece out of range in SomaSolution::get()" << std::endl;
    return false;
  }
#endif
  
  if(orientation[piece] == SomaPiece::ORIENTATIONS) return false;
  
  *pos = position[piece];
  *ori = orientation[piece];

  return true;
}

//**************************************************************************  

void SomaSolution::write(std::ostream& strm)const{
  strm << " {";
  strm << " (";
  strm.width(3);
  strm << num_pieces;
  strm << ")";
  for(unsigned int i=0;i<num_pieces;i++){
    strm << " (";
    strm.width(3);
    strm << orientation[i];
    strm << ")";
    strm << position[i];
    if(!((i+1) % 3)) strm << std::endl;
  }
  strm << " }" << std::endl;
}

//**************************************************************************  

bool SomaSolution::read(std::istream& strm){
  char S=' ';
  
  S = ' ';
  strm >> S;
  if(S != '{'){
    std::cerr << "Error in SomaSolution::read(). '{' expected." << std::endl;
    return false;
  } 
  
  
  S = ' ';
  strm >> S;
  if(S != '('){
    std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
    return false;
  }
  
  unsigned int new_num_pieces=0;
  strm >> new_num_pieces;
  if(new_num_pieces == 0){
    std::cerr << "Error in SomaSolution::read(). Number of pieces 0." 
              << std::endl;
    return false;  
  }
  
  S = ' ';
  strm >> S;
  if(S != ')'){
    std::cerr << "Error in SomaSolution::read(). ')' expected." << std::endl;
    return false;
  }
  
  num_pieces = new_num_pieces;
  orientation.resize(num_pieces);
  position.resize(num_pieces);
  
  for(unsigned int i=0;i<num_pieces;i++){
    S = ' ';
    strm >> S;
    if(S != '('){
      std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
      return false;
    }else{
      strm >> orientation[i];
      if(orientation[i] > SomaPiece::ORIENTATIONS){
        std::cerr << "Error in SomaSolution::read(). Orientation "
                  << orientation[i] << " out of range." << std::endl;
        return false;
      }
    }
    
    S = ' ';
    strm >> S;
    if(S != ')'){
      std::cerr << "Error in SomaSolution::read(). '(' expected." << std::endl;
      return false;  
    }else{
      if(! position[i].read(strm)) return false;
    }
    
  }
  
  S = ' ';
  strm >> S;
  if(S != '}'){
    std::cerr << "Error in SomaSolution::read(). '}' expected." << std::endl;
    return false;
  } 
  
  return true;
}

//**************************************************************************
//**************************************************************************
//**************************************************************************

SomaSolutions::SomaSolutions(){}
  
//**************************************************************************

SomaSolutions::~SomaSolutions(){
  for(unsigned int i=0;i<solutions.size();i++){
    delete solutions[i];
  }
}
  
//**************************************************************************

void SomaSolutions::addSolution(const SomaSolution& solution){
  solutions.push_back(new SomaSolution(solution));
}
  
//**************************************************************************

const SomaSolution* SomaSolutions::get(const unsigned int number)const{
#ifdef SOMA_DEBUG
 if(number >= solutions.size()){
   std::cerr << "Number out of range in SomaSolutions::get()" << std::endl;
   return 0;
 }
#endif

 return solutions[number];
}
  
//**************************************************************************

SomaSolution* SomaSolutions::get(const unsigned int number){
#ifdef SOMA_DEBUG
 if(number >= solutions.size()){
   std::cerr << "Number out of range in SomaSolutions::get()" << std::endl;
   return 0;
 }
#endif

 return solutions[number];
}
  
//**************************************************************************

void SomaSolutions::write(std::ostream& strm)const{
  strm << " (";
  strm.width(3);
  strm << solutions.size();
  strm << ")" << std::endl;
  for(unsigned int i=0;i<solutions.size();i++){
    solutions[i]->write(strm);
  }
}
  
//**************************************************************************

bool SomaSolutions::read(std::istream& strm){
  char S=' ';
  
  unsigned int num_solutions = 0;
  std::vector<SomaSolution* > readsolutions;
  
  strm >> S;
  if(S != '('){
    std::cerr << "Error in SomaSolutions::read(). '(' expected." << std::endl;
    return false;
  }
  
  strm >> num_solutions;
  if(num_solutions == 0){
    std::cerr << "Error in SomaSolutions::read(). Number of solutions is 0." 
              << std::endl;
    return false;
  }
  
  S = ' ';
  strm >> S;
  if(S != ')'){
    std::cerr << "Error in SomaSolutions::read(). ')' expected." << std::endl;
    return false;
  }
  
  bool noerror = true;
  for(unsigned int i=0;i<num_solutions;i++){
    SomaSolution *readsolution = new SomaSolution(1);
    noerror = readsolution->read(strm);
    readsolutions.push_back(readsolution);
    if(!noerror) break;
  }
  
  if(!noerror){
    for(unsigned int i=0;i<readsolutions.size();i++){
      delete readsolutions[i];
    }
    return false;
  }else{
    for(unsigned int i=0;i<solutions.size();i++){
      delete solutions[i];
    }
    solutions.clear();
    for(unsigned int i=0;i<readsolutions.size();i++){
      solutions.push_back(readsolutions[i]);
    }
    return true;
  }

}
  
//**************************************************************************

